// version 1.0.0
unit GDDMatrix;

interface

uses Math, classes;

type TGDDMatrix= class
   public
    Data:array of array of double;
    NbLignes:Integer;
    NbColonnes:Integer;
  function Norme:double;
  function Add(var Mat:TGDDMatrix):boolean;
  function Multiply (var MatB, MatRes : TGDDMatrix): boolean;
  function LUDecompositionPivot(var MatL, MatU, Matr:TGDDMatrix):integer;
  function Invert: boolean;
  function MaxEigenValVec(var EigenValue : double;var EigenVector : TGDDMatrix; eps:double):boolean ;
  procedure SMult(scalar : double);
  procedure Clone (var MatSource: TGDDMatrix);
  procedure Transpose();
  constructor Create(NbLig, NbCol:Integer); overload;
  constructor Create(); overload;
  procedure Initialisation(NbLig, NbCol:Integer);
  function InvertTrianInf():boolean;
  function InvertTrianSup():boolean;
end;

implementation

constructor TGDDMatrix.Create(NbLig, NbCol:Integer);
begin
  NbLignes := NbLig;
  NbColonnes := NbCol;
  setLength(Data,NbLignes,NbColonnes);
end;

constructor TGDDMatrix.Create();
begin
  Create(0,0);
end;


procedure TGDDMatrix.Initialisation(NbLig, NbCol:Integer);
begin
  NbLignes := NbLig;
  NbColonnes := NbCol;
  SetLength(Data,0,0);
  SetLength(Data,NbLignes,NbColonnes);
end;

// Norme euclidienne d'une matrice
function TGDDMatrix.Norme:double;
var
  i,j:Integer;
begin
	result:=0;
	for i:=0 to NbLignes-1 do
  for j:=0 to NbColonnes-1 do
      result := result +  sqr(Data[i,j]);
  result := sqrt(result);
end;

// Addition de 2 matrices
function TGDDMatrix.Add(var Mat:TGDDMatrix):boolean;
var
  i,j:Integer;
begin
  if (Mat.NbLignes=NbLignes) and (Mat.NbColonnes=NbColonnes) then
    result := true else result:=false;
  if result then
    begin
      for i:=0 to NbLignes-1 do
      for j:=0 to NbColonnes-1 do
        Data[i,j] := Data[i,j] + Mat.Data[i,j];
    end;
end;

// Multiplication de 2 matrices
function TGDDMatrix.Multiply (var MatB, MatRes : TGDDMatrix): boolean;
var
  i,j,k:Integer;
begin
  if NbColonnes = MatB.NbLignes then result := true else result := false;
  if result then
    begin
    MatRes.Initialisation(NbLignes,MatB.NbColonnes);
    for i:=0 to MatRes.NbLignes-1 do
    for j:=0 to MatRes.NbColonnes-1 do
        for k:=0 to NbColonnes-1 do
          MatRes.Data[i,j] := MatRes.data[i,j] + Data[i,k] * MatB.Data[k,j];
    end;
end;


// Multiplication d'une matrice par un scalaire
procedure TGDDMatrix.SMult(scalar : double);
var
  i,j:Integer;
begin
  for i:=0 to NbLignes-1 do
  for j:=0 to NbColonnes-1 do
    Data[i,j] := Data[i,j] * scalar;
end;

// Copie physique d'une matrice dans une autre
procedure TGDDMatrix.Clone(var MatSource: TGDDMatrix);
var
  i,j:Integer;
begin
  Initialisation(MatSource.NbLignes,MatSource.NbColonnes);
  for i:=0 to NbLignes-1 do
  for j:=0 to NbColonnes-1 do
    Data[i,j] := MatSource.Data[i,j];
end;

// Retourne la transpose d'une matrice
procedure TGDDMatrix.Transpose();
var
  i,j:Integer;
  MatTemp:TGDDMatrix;
begin
  MatTemp:=TGDDMatrix.create(NbColonnes,NbLignes);
  for i:=0 to NbLignes-1 do
  for j:=0 to NbColonnes-1 do
    MatTemp.Data[j][i] := Data[i][j];
  Clone(MatTemp);
  MatTemp.Free;
end;

// Calcul la dcomposition LU d'une matrice
function TGDDMatrix.LUDecompositionPivot(var MatL, MatU, Matr:TGDDMatrix):integer;
var
  i,j,k,rr,err:Integer;
  max,pivot,coef:double;
  r:array of Integer;
begin
 MatL.Initialisation(NbLignes,NbColonnes);
 MatU.Initialisation(NbLignes,NbColonnes);
 Matr.Initialisation(NbLignes,NbColonnes);
 setlength(r,NbLignes);

 for i:=0 to NbLignes-1 do
  begin
  r[i] :=i ;
  for j:=0 to NbLignes-1 do  MatU.data[i,j] := Data[i,j];
  end;


 err :=1;
 k :=1;
 while (err=1) And (k<NbLignes) do
  begin
  max := abs(MatU.Data[k-1,k-1]);
  j := k;
  for i:=k+1 to NbLignes do
    if(max<abs(MatU.Data[i-1,k-1])) then
      begin
       max:=abs(MatU.Data[i-1,k-1]);
       j:=i;
      end;
  if(max<>0) then
  begin
   if (j<>k) then
    begin
    rr :=r[k-1];
    r[k-1]:=r[j-1];
    r[j-1]:=rr;
    for i:=1 to k-1 do
      begin
       coef:=MatL.Data[k-1,i-1];
       MatL.Data[k-1,i-1]:=MatL.Data[j-1,i-1];
       MatL.Data[j-1,i-1]:=coef;
      end;

    for i:=k to NbLignes do
      begin
       pivot:=MatU.Data[k-1,i-1];
       MatU.Data[k-1,i-1]:=MatU.Data[j-1,i-1];
       MatU.Data[j-1,i-1]:=pivot;
      end;
  end;

   pivot:=MatU.Data[k-1,k-1];
   for i:=k+1 to NbLignes do
    begin
      coef:=MatU.Data[i-1,k-1]/pivot;
      for j:=k+1 to NbLignes do
        MatU.Data[i-1,j-1] := MatU.Data[i-1,j-1] - coef*MatU.Data[k-1,j-1];
      MatL.Data[i-1,k-1]:=coef;
   end;
  end
  else err:=0;
  inc(k);
 end;
 if(MatU.Data[NbLignes-1,NbLignes-1]=0) then err:=0;

   for i:=0 to NbLignes-1 do
       for j:=0 to i-1 do MatU.Data[i,j]:=0;
   for i:=0 to NbLignes-1 do MatL.Data[i,i]:=1;
   for i:=0 to NbLignes-1 do Matr.Data[i,r[i]]:=1;
 result := err;
end;


// Retourne l'inverse d'une matrice  partir d'une dcomposition LU
function TGDDMatrix.Invert: boolean;
var
  MatU,MatL,Matr:TGDDMatrix;
  MatTemp:TGDDMatrix;
begin
  MatU:=TGDDMatrix.Create();
  MatL:=TGDDMatrix.Create();
  Matr:=TGDDMatrix.Create();
  Mattemp:=TGDDMatrix.Create();
  try;
  LUDecompositionPivot(MatL,MatU,Matr);
  result := MatL.InvertTrianInf();
  if result then result := MatU.InvertTrianSup() else exit;
  if result then MatU.Multiply(MatL,MatTemp) else exit;
  Mattemp.Multiply(Matr,Self);
  finally
  MatU.Free;
  MatL.Free;
  Mattemp.Free;
  Matr.Free;
  end;
end;

// Retourne l'inverse d'une matrice triangulaire infrieure
function TGDDMatrix.InvertTrianInf():boolean;
var
  i,j,k:Integer;
  MatInvert:TGDDMatrix;
begin
  try
  MatInvert := TGDDMatrix.Create(NbLignes,NbColonnes);
  if NbLignes<>NbColonnes then begin result:=false; exit; end;
  for i:=0 to NbLignes-1 do
      if Data[i][i]<>0 then MatInvert.Data[i][i] := 1/Data[i][i]
        else
        begin
        result:=false;
        exit;
        end;
  result:= true;

  for i:=NbLignes-1 downto 0 do
  for j:=NbColonnes-1 downto i+1 do
    begin
			for k:=j downto i+1 do
				MatInvert.Data[j][i] := MatInvert.Data[j][i] - Data[k][i]*MatInvert.Data[j][k];
			MatInvert.Data[j][i] := MatInvert.Data[j][i] / Data[i][i];
    end;
  Clone(MatInvert);
  finally
  MatInvert.Free;
  end;
end;

// Retourne l'inverse d'une matrice triangulaire suprieure
function TGDDMatrix.InvertTrianSup():boolean;
var
  i,j,k:Integer;
  MatInvert:TGDDMatrix;
begin
  try
  MatInvert := TGDDMatrix.Create(NbLignes,NbColonnes);
  if NbLignes<>NbColonnes then begin result:=false; exit; end;
  for i:=0 to NbLignes-1 do
      if Data[i][i]<>0 then MatInvert.Data[i][i] := 1/Data[i][i]
        else
        begin
        result:=false;
        exit;
        end;
  result:= true;
   for i:=NbLignes-1 downto 0 do
   for j:=NbColonnes-1 downto i+1 do
    begin
			for k:=j downto i+1 do
				MatInvert.Data[i][j] := MatInvert.Data[i][j] - Data[i][k]*MatInvert.Data[k][j];
			MatInvert.Data[i][j] := MatInvert.Data[i][j] / Data[i][i];
    end;
 Clone(MatInvert);
  finally
  MatInvert.Free;
  end;
end;

// Retourne la plus grande valeur propre et le vecteur propre associe, eps tant la prcision souhaite
// sur la valeur propre
function TGDDMatrix.MaxEigenValVec(var EigenValue : double;var EigenVector : TGDDMatrix; eps:double):boolean;
var
  i,c:Integer;
  EigenValueTemp,lanorme:double;
  EigenVectorTemp : TGDDMatrix;
  epsencours:double;
  muleps:double;
  Compteur :Integer;
begin
  EigenVectorTemp := TGDDMatrix.Create(NbColonnes,1);
  EigenVector.Initialisation(NbLignes,1);
  muleps:=10e-10;
  epsencours := eps;
  EigenVector.Data[0,0] := 1;
  EigenValue:=0;
  Compteur := 0;
  EigenValueTemp :=eps+1;
  while abs(EigenValue-EigenValueTemp)>epsencours do
  begin
    c := 1;
    inc(Compteur);
    if Compteur > 100000 then
      begin
      if eps=0 then
        begin
        muleps := muleps *10;
        if muleps>10e-5 then begin result:=false; exit; end;
        end
        else
        begin
        result:=false;
        exit;
        end;
      end;
    EigenValueTemp := EigenValue;
    Multiply(Eigenvector,EigenVectorTemp);
    while EigenVectorTemp.Data[c][0]=0 do inc(c);
    EigenValue := EigenVectorTemp.Data[c][0];
    for i:=0 to NbColonnes-1 do
        Eigenvector.Data[i][0] := EigenVectorTemp.data[i][0] / EigenValue;
    if (eps=0) then epsencours := abs(EigenValue * muleps);
  end;

  result:=true;
  lanorme := Eigenvector.Norme;
  for i:=0 to Eigenvector.NbLignes-1 do
      Eigenvector.Data[i][0] := Eigenvector.Data[i][0] / lanorme;
  EigenVectorTemp.Free;
end;

end.


